import inspect
from concurrent.futures import ThreadPoolExecutor
from typing import Annotated, Iterator, Union

from fastapi import APIRouter, Depends, FastAPI
from starlette.requests import Request
from typing_extensions import TypeAlias

from ncae_sdk._async.client import AsyncClient
from ncae_sdk._sync.client import Client
from ncae_sdk._types import ParamProcessorContext, _extract_class_type


class DependencyMixin:
    pass


def annotate_fastapi_dependencies(ctx: ParamProcessorContext) -> Iterator[inspect.Parameter]:
    for param in ctx.current:
        original_param = ctx.original_map[param.name]
        param_class = _extract_class_type(original_param.annotation)
        if param_class and issubclass(param_class, DependencyMixin):
            yield param.replace(annotation=Annotated[param.annotation, Depends()])
        else:
            yield param


def get_thread_pool(request: Request) -> ThreadPoolExecutor:
    thread_pool = getattr(request.state, "thread_pool", None)
    if not isinstance(thread_pool, ThreadPoolExecutor):
        raise RuntimeError("Thread pool not found in request state.")

    return thread_pool


def get_ncae_client_sync(request: Request) -> Client:
    sync_client = getattr(request.state, "ncae_client_sync")
    if not isinstance(sync_client, Client):
        raise RuntimeError("NcaeClient not found in request state.")

    return sync_client


def get_ncae_client_async(request: Request) -> AsyncClient:
    async_client = getattr(request.state, "ncae_client_async")
    if not isinstance(async_client, AsyncClient):
        raise RuntimeError("NcaeClient not found in request state.")

    return async_client


ThreadPoolExecutorDep: TypeAlias = Annotated[ThreadPoolExecutor, Depends(get_thread_pool)]
NcaeClientDep: TypeAlias = Annotated[Client, Depends(get_ncae_client_sync)]
AsyncNcaeClientDep: TypeAlias = Annotated[AsyncClient, Depends(get_ncae_client_async)]
Router: TypeAlias = Union[FastAPI, APIRouter]



import os
from functools import cached_property
from types import TracebackType
from typing import Optional, Type

from httpx import AsyncBaseTransport
from typing_extensions import Self

from ncae_sdk._async.session import AsyncSession
from ncae_sdk._auth import Auth, SessionAuth
from ncae_sdk._session import SessionContext
from ncae_sdk._util import parse_env_bool, parse_env_int
from ncae_sdk.endpoints import (
    AsyncAuditLogEndpoint,
    AsyncAwxModuleEndpoint,
    AsyncBasePhaseEndpoint,
    AsyncChangeScheduleEndpoint,
    AsyncChangeScheduleRunEndpoint,
    AsyncCmdbEntryEndpoint,
    AsyncCredentialEndpoint,
    AsyncDeviceEndpoint,
    AsyncDeviceGroupEndpoint,
    AsyncDeviceModelEndpoint,
    AsyncExtApiAuthEndpoint,
    AsyncExtApiPhaseEndpoint,
    AsyncExtApiServiceEndpoint,
    AsyncGrafanaDashboardEndpoint,
    AsyncGrafanaSystemEndpoint,
    AsyncInfluxRepositoryEndpoint,
    AsyncInfluxSystemEndpoint,
    AsyncModuleEndpoint,
    AsyncPhaseEndpoint,
    AsyncPhaseInstanceEndpoint,
    AsyncPollingPhaseEndpoint,
    AsyncPythonScriptPhaseEndpoint,
    AsyncServiceEndpoint,
    AsyncServiceInstanceEndpoint,
    AsyncServiceInstanceLogEndpoint,
    AsyncServiceManagementPhaseEndpoint,
    AsyncSharedDataBucketEndpoint,
    AsyncSharedDataSubscriberEndpoint,
    AsyncTagEndpoint,
    AsyncTagRelationEndpoint,
    AsyncTenantEndpoint,
    AsyncTransactionEndpoint,
    AsyncUploadFileEndpoint,
    AsyncUsageLogEndpoint,
    AsyncUserEndpoint,
)
from ncae_sdk.endpoints._async.favourite import AsyncFavouriteEndpoint
from ncae_sdk.endpoints._async.report import AsyncReportEndpoint


class AsyncClient:
    def __init__(
        self,
        base_url: Optional[str] = None,
        auth: Optional[Auth] = None,
        timeout: Optional[int] = None,
        verify: Optional[bool] = None,
        httpx_transport: Optional[AsyncBaseTransport] = None,
    ) -> None:
        """
        Initializes a new asynchronous client for interacting with the NCAE API.

        This client provides access to various endpoints such as audit logs, devices, credentials, and more. It is
        based on an underlying session that uses httpx for asynchronous HTTP requests. For these reasons, it is
        recommended to use this client as a context manager, as this ensures cleanup and allows proper pooling.

        All arguments are marked as optional, which will fallback to environment variables or default values.
        If no suitable configuration is found, an error will be raised during instantiation.

        :param base_url: The base URL of the NCAE API, e.g. `https://ncae.example.com/api/`.
        :param auth: An instance of :class:`Auth` that provides authentication for the API requests.
        :param timeout: The desired timeout for API requests in seconds. Usage of default value is recommended.
        :param verify: Whether to verify SSL certificates. Set to `False` if you are using a self-signed certificate.
        """
        if not base_url:
            base_url = os.environ.get("NCAE_SDK_BASE_URL")
            if not base_url:
                raise ValueError("Unable to determine NCAE SDK base URL.")

        if not auth:
            username = os.environ.get("NCAE_SDK_USERNAME")
            password = os.environ.get("NCAE_SDK_PASSWORD")
            if not username or not password:
                raise ValueError("Unable to determine NCAE SDK authentication credentials.")

            auth = SessionAuth(username=username, password=password)

        if timeout is None:
            if value := os.environ.get("NCAE_SDK_HTTP_TIMEOUT"):
                timeout = parse_env_int(value)
            else:
                timeout = 5

        if verify is None:
            if value := os.environ.get("NCAE_SDK_HTTP_VERIFY"):
                verify = parse_env_bool(value)
            else:
                verify = True

        self._session: AsyncSession = AsyncSession.create(
            base_url=base_url,
            auth=auth,
            timeout=timeout,
            verify=verify,
            httpx_transport=httpx_transport,
        )

    def clone_with_context(self, context: SessionContext) -> Self:
        instance = self.__class__.__new__(self.__class__)
        instance._session = self._session.clone(context)
        return instance

    async def close(self) -> None:
        await self._session.close()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def ping(self) -> None:
        await self._session.get("readyz")

    @property
    def session(self) -> AsyncSession:
        return self._session

    @cached_property
    def audit_logs(self) -> AsyncAuditLogEndpoint:
        return AsyncAuditLogEndpoint(session=self._session)

    @cached_property
    def awx_modules(self) -> AsyncAwxModuleEndpoint:
        return AsyncAwxModuleEndpoint(session=self._session)

    @cached_property
    def base_phases(self) -> AsyncBasePhaseEndpoint:
        return AsyncBasePhaseEndpoint(session=self._session)

    @cached_property
    def change_schedules(self) -> AsyncChangeScheduleEndpoint:
        return AsyncChangeScheduleEndpoint(session=self._session)

    @cached_property
    def change_schedule_runs(self) -> AsyncChangeScheduleRunEndpoint:
        return AsyncChangeScheduleRunEndpoint(session=self._session)

    @cached_property
    def cmdb_entries(self) -> AsyncCmdbEntryEndpoint:
        return AsyncCmdbEntryEndpoint(session=self._session)

    @cached_property
    def credentials(self) -> AsyncCredentialEndpoint:
        return AsyncCredentialEndpoint(session=self._session)

    @cached_property
    def devices(self) -> AsyncDeviceEndpoint:
        return AsyncDeviceEndpoint(session=self._session)

    @cached_property
    def device_groups(self) -> AsyncDeviceGroupEndpoint:
        return AsyncDeviceGroupEndpoint(session=self._session)

    @cached_property
    def device_models(self) -> AsyncDeviceModelEndpoint:
        return AsyncDeviceModelEndpoint(session=self._session)

    @cached_property
    def ext_api_auths(self) -> AsyncExtApiAuthEndpoint:
        return AsyncExtApiAuthEndpoint(session=self._session)

    @cached_property
    def ext_api_services(self) -> AsyncExtApiServiceEndpoint:
        return AsyncExtApiServiceEndpoint(session=self._session)

    @cached_property
    def ext_api_phases(self) -> AsyncExtApiPhaseEndpoint:
        return AsyncExtApiPhaseEndpoint(session=self._session)

    @cached_property
    def favourites(self) -> AsyncFavouriteEndpoint:
        return AsyncFavouriteEndpoint(session=self._session)

    @cached_property
    def grafana_dashboards(self) -> AsyncGrafanaDashboardEndpoint:
        return AsyncGrafanaDashboardEndpoint(session=self._session)

    @cached_property
    def grafana_systems(self) -> AsyncGrafanaSystemEndpoint:  # noqa: F821
        return AsyncGrafanaSystemEndpoint(session=self._session)

    @cached_property
    def influx_repositories(self) -> AsyncInfluxRepositoryEndpoint:
        return AsyncInfluxRepositoryEndpoint(session=self._session)

    @cached_property
    def influx_systems(self) -> AsyncInfluxSystemEndpoint:
        return AsyncInfluxSystemEndpoint(session=self._session)

    @cached_property
    def modules(self) -> AsyncModuleEndpoint:
        return AsyncModuleEndpoint(session=self._session)

    @cached_property
    def phases(self) -> AsyncPhaseEndpoint:
        return AsyncPhaseEndpoint(session=self._session)

    @cached_property
    def phase_instances(self) -> AsyncPhaseInstanceEndpoint:
        return AsyncPhaseInstanceEndpoint(session=self._session)

    @cached_property
    def polling_phases(self) -> AsyncPollingPhaseEndpoint:
        return AsyncPollingPhaseEndpoint(session=self._session)

    @cached_property
    def python_script_phases(self) -> AsyncPythonScriptPhaseEndpoint:
        return AsyncPythonScriptPhaseEndpoint(session=self._session)

    @cached_property
    def reports(self) -> AsyncReportEndpoint:
        return AsyncReportEndpoint(session=self._session)

    @cached_property
    def services(self) -> AsyncServiceEndpoint:
        return AsyncServiceEndpoint(session=self._session)

    @cached_property
    def service_instances(self) -> AsyncServiceInstanceEndpoint:
        return AsyncServiceInstanceEndpoint(session=self._session)

    @cached_property
    def service_instance_logs(self) -> AsyncServiceInstanceLogEndpoint:
        return AsyncServiceInstanceLogEndpoint(session=self._session)

    @cached_property
    def shared_data_buckets(self) -> AsyncSharedDataBucketEndpoint:
        return AsyncSharedDataBucketEndpoint(session=self._session)

    @cached_property
    def shared_data_subscribers(self) -> AsyncSharedDataSubscriberEndpoint:
        return AsyncSharedDataSubscriberEndpoint(session=self._session)

    @cached_property
    def service_management_phases(self) -> AsyncServiceManagementPhaseEndpoint:
        return AsyncServiceManagementPhaseEndpoint(session=self._session)

    @cached_property
    def tags(self) -> AsyncTagEndpoint:
        return AsyncTagEndpoint(session=self._session)

    @cached_property
    def tag_relations(self) -> AsyncTagRelationEndpoint:
        return AsyncTagRelationEndpoint(session=self._session)

    @cached_property
    def tenants(self) -> AsyncTenantEndpoint:
        return AsyncTenantEndpoint(session=self._session)

    @cached_property
    def transactions(self) -> AsyncTransactionEndpoint:
        return AsyncTransactionEndpoint(session=self._session)

    @cached_property
    def upload_files(self) -> AsyncUploadFileEndpoint:
        return AsyncUploadFileEndpoint(session=self._session)

    @cached_property
    def usage_logs(self) -> AsyncUsageLogEndpoint:
        return AsyncUsageLogEndpoint(session=self._session)

    @cached_property
    def users(self) -> AsyncUserEndpoint:
        return AsyncUserEndpoint(session=self._session)



import logging
from typing import Union, cast

from fastapi import Request
from pydantic import AliasChoices, AliasPath, BaseModel
from pydantic.fields import FieldInfo

from ncae_sdk.fastapi._logging import ContextLogAdapter, ContextLogger


def extract_pydantic_model_keys(
    model_cls: type[BaseModel],
    *,
    validate_by_name: bool = False,
    validate_by_alias: bool = True,
) -> set[str]:
    result: set[str] = set()
    for field_name, field in model_cls.model_fields.items():
        assert isinstance(field, FieldInfo)
        field_aliases = _extract_pydantic_field_keys(field.validation_alias)

        if validate_by_name:
            result.add(field_name)
        if validate_by_alias:
            result.update(field_aliases)

    return result


def _extract_pydantic_field_keys(value: Union[str, int, AliasPath, AliasChoices, None]) -> set[str]:
    if value is None:
        return set()

    if isinstance(value, (str, int)):
        return {str(value)}
    elif isinstance(value, AliasPath):
        return {str(value.path[0])}
    elif isinstance(value, AliasChoices):
        result: set[str] = set()
        for choice in value.choices:
            result.update(_extract_pydantic_field_keys(choice))

        return result
    else:
        raise TypeError(f"Unsupported alias type: {type(value)}")


def obtain_request_logger(request: Request, recreate: bool = False) -> ContextLogger:
    # Use existing logger from request state if available and not recreating
    logger = getattr(request.state, "logger", None)
    if isinstance(logger, (logging.Logger, logging.LoggerAdapter)) and not recreate:
        delattr(request.state, "logger")
        return logger

    # Obtain correlation ID, preferably from request state set by middleware, but falling back to headers
    # This is required in some cases as e.g. exception handlers do not pass through middleware
    base_logger = logging.getLogger("ncae_sdk.request")
    correlation_id = getattr(request.state, "correlation_id", None) or request.headers.get("X-Correlation-Id", None)
    if not correlation_id:
        return base_logger

    # Create and store a new ContextLogAdapter in request state
    request.state.logger = ContextLogAdapter(
        logger=base_logger,
        extra={"identifier": correlation_id},
    )

    return cast(ContextLogAdapter, request.state.logger)



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

from collections.abc import Iterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._resource import ResourceId
from ncae_sdk._sync.endpoint import Endpoint
from ncae_sdk._util import map_sync
from ncae_sdk.resources._schema import (
    InfluxSystem,
    InfluxSystemCreate,
    InfluxSystemCreateModel,
    InfluxSystemFilter,
    InfluxSystemFilterModel,
    InfluxSystemUpdate,
    InfluxSystemUpdateModel,
)


class InfluxSystemEndpoint(Endpoint):
    BASE_PATH: Final[str] = "observability/v1/influx-system"

    def list(self, **filters: Unpack[InfluxSystemFilter]) -> Iterator[InfluxSystem]:
        results = self._list(self.BASE_PATH, InfluxSystemFilterModel, filters)
        return map_sync(InfluxSystem.parse_api, results)

    def find(self, **filters: Unpack[InfluxSystemFilter]) -> Optional[InfluxSystem]:
        result = self._get_by_filters(self.BASE_PATH, InfluxSystemFilterModel, filters)
        return InfluxSystem.parse_api(result) if result else None

    def get(self, rid: ResourceId) -> Optional[InfluxSystem]:
        result = self._get_by_id(self.BASE_PATH, rid)
        return InfluxSystem.parse_api(result) if result else None

    def create(self, **payload: Unpack[InfluxSystemCreate]) -> InfluxSystem:
        result = self._create(self.BASE_PATH, InfluxSystemCreateModel, payload)
        return InfluxSystem.parse_api(result)

    def update(self, rid: ResourceId, /, **payload: Unpack[InfluxSystemUpdate]) -> InfluxSystem:
        result = self._update(self.BASE_PATH, rid, InfluxSystemUpdateModel, payload)
        return InfluxSystem.parse_api(result)

    def delete(self, rid: ResourceId) -> bool:
        return self._delete(self.BASE_PATH, rid)



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    CmdbEntry,
    CmdbEntryCreate,
    CmdbEntryCreateModel,
    CmdbEntryFilter,
    CmdbEntryFilterModel,
    CmdbEntryUpdate,
    CmdbEntryUpdateModel,
)


class AsyncCmdbEntryEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "cmdb/v1/entry"

    async def list(self, **filters: Unpack[CmdbEntryFilter]) -> AsyncIterator[CmdbEntry]:
        results = self._list(self.BASE_PATH, CmdbEntryFilterModel, filters)
        return map_async(CmdbEntry.parse_api, results)

    async def find(self, **filters: Unpack[CmdbEntryFilter]) -> Optional[CmdbEntry]:
        result = await self._get_by_filters(self.BASE_PATH, CmdbEntryFilterModel, filters)
        return CmdbEntry.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[CmdbEntry]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return CmdbEntry.parse_api(result) if result else None

    async def create(self, **payload: Unpack[CmdbEntryCreate]) -> CmdbEntry:
        result = await self._create(self.BASE_PATH, CmdbEntryCreateModel, payload)
        return CmdbEntry.parse_api(result)

    async def update(self, rid: ResourceId, /, **payload: Unpack[CmdbEntryUpdate]) -> CmdbEntry:
        result = await self._update(self.BASE_PATH, rid, CmdbEntryUpdateModel, payload)
        return CmdbEntry.parse_api(result)

    async def delete(self, rid: ResourceId) -> bool:
        return await self._delete(self.BASE_PATH, rid)



import functools
import inspect
from typing import Callable, Final, Generic, Iterator, Optional, TypeVar

from typing_extensions import Concatenate, ParamSpec

from ncae_sdk.fastapi.runner._progress import ProgressHookAsync, ProgressHookSync, ProgressTracker, ProgressUpdate
from ncae_sdk.fastapi.runner._utils import ContextT, ItemT, RunnerResult

P = ParamSpec("P")
R = TypeVar("R")


class ParallelRunner(Generic[ContextT, ItemT]):
    def __init__(
        self,
        ctx: ContextT,
        items: list[ItemT],
        hook_sync: ProgressHookSync,
        hook_async: ProgressHookAsync,
    ) -> None:
        self._ctx: Final[ContextT] = ctx
        self._items: Final[list[ItemT]] = items
        self._progress: Final[ProgressTracker] = ProgressTracker(len(self._items))
        self._hook_sync: Final[ProgressHookSync] = hook_sync
        self._hook_async: Final[ProgressHookAsync] = hook_async

    @property
    def progress(self) -> ProgressTracker:
        return self._progress

    def run_sync(
        self,
        func: Callable[Concatenate[ContextT, ItemT, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iterator[RunnerResult[ItemT, R]]:
        if not callable(func) or inspect.iscoroutinefunction(func):
            raise TypeError("The provided function must be a synchronous function.")

        @functools.wraps(func)
        def wrapper(item: ItemT) -> RunnerResult[ItemT, R]:
            update: Optional[ProgressUpdate] = None
            try:
                result = func(self._ctx, item, *args, **kwargs)
                update = self._progress.increment("succeeded")
                return RunnerResult(item=item, exception=None, raw_value=result)
            except Exception as exc:
                update = self._progress.increment("failed")
                return RunnerResult(item=item, exception=exc, raw_value=None)
            finally:
                try:
                    if update is not None:
                        self._hook_sync(update)
                except Exception as exc:
                    self._ctx.internal_logger.warning("Execution of progress hook failed: %s", exc)

        self._progress.reset()
        futures = [self._ctx.thread_pool.submit(wrapper, item) for item in self._items]
        for future in futures:
            yield future.result()

    def _process_item_sync(
        self,
        func: Callable[Concatenate[ContextT, ItemT, P], R],
        item: ItemT,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RunnerResult[ItemT, R]:
        update: Optional[ProgressUpdate] = None
        try:
            result = func(self._ctx, item, *args, **kwargs)
            update = self._progress.increment("succeeded")
            return RunnerResult(item=item, exception=None, raw_value=result)
        except Exception as exc:
            update = self._progress.increment("failed")
            return RunnerResult(item=item, exception=exc, raw_value=None)
        finally:
            try:
                if update is not None:
                    self._hook_sync(update)
            except Exception as exc:
                self._ctx.internal_logger.warning("Execution of progress hook failed: %s", exc)



# @generated
# This file was automatically generated by the ncae-sdk code generator.
# DO NOT EDIT MANUALLY, ANY CHANGES WILL BE OVERWRITTEN.

import socket
from typing import Any, Generic, Optional

from ncae_sdk.fastapi._context import BaseContext
from ncae_sdk.fastapi.extapi._contracts import CallbackT, CmdbT
from ncae_sdk.fastapi.extapi._protocols import ExtApiPhaseContextProtocol
from ncae_sdk.fastapi.runner._progress import ProgressUpdate
from ncae_sdk.types import LogLevel, PhaseStatus


class ExtApiPhaseContextMixin(BaseContext, ExtApiPhaseContextProtocol[CmdbT, CallbackT], Generic[CmdbT, CallbackT]):
    def log_debug(self, message: str, *args: Any, title: Optional[str] = None, hostname: Optional[str] = None) -> None:
        self.log(LogLevel.DEBUG, message, *args, title=title, hostname=hostname)

    def log_info(self, message: str, *args: Any, title: Optional[str] = None, hostname: Optional[str] = None) -> None:
        self.log(LogLevel.INFO, message, *args, title=title, hostname=hostname)

    def log_warning(
        self, message: str, *args: Any, title: Optional[str] = None, hostname: Optional[str] = None
    ) -> None:
        self.log(LogLevel.WARNING, message, *args, title=title, hostname=hostname)

    def log_error(self, message: str, *args: Any, title: Optional[str] = None, hostname: Optional[str] = None) -> None:
        self.log(LogLevel.ERROR, message, *args, title=title, hostname=hostname)

    def log(
        self,
        level: LogLevel,
        message: str,
        *args: Any,
        title: Optional[str] = None,
        hostname: Optional[str] = None,
    ) -> None:
        message = (message % args) if args else message
        title = title or self._request.app.title
        hostname = hostname or socket.gethostname()

        self.internal_logger.log(
            level.to_native(),
            "Publishing service [%(hostname)s] %(title)s: %(message)s",
            {"title": title, "hostname": hostname, "message": message},
        )

        self.client.service_instance_logs.create(
            status=level,
            hostname=hostname or socket.gethostname(),
            title=title,
            text=message,
            service_instance_id=self.body.ncae_service_instance_id,
        )

    def get_phase_status(self) -> PhaseStatus:
        phase = self.client.phase_instances.get(self.body.ncae_phase_instance_id)
        assert phase is not None, "Could not retrieve own phase instance"

        return phase.status

    def update_phase_status(self, status: PhaseStatus) -> None:
        self.client.phase_instances.update(self.body.ncae_phase_instance_id, status=status)

    def update_callback_data(self) -> None:
        callback_data = self.callback_data.model_dump(mode="json")
        self.client.cmdb_entries.update(
            self.body.ncae_cmdb_entry_id,
            callback_data=callback_data,
        )

    def _publish_progress_update(self, update: ProgressUpdate) -> None:
        self.log_info(
            "Processed %d out of %d items in %.2f seconds (%.2f%%)",
            update.current,
            update.total,
            update.elapsed_time.total_seconds(),
            update.percentage,
            title=f"Progress Update ({update.percentage:.2f}%)",
        )



from collections.abc import AsyncIterator
from typing import Final, Optional

from typing_extensions import Unpack

from ncae_sdk._async.endpoint import AsyncEndpoint
from ncae_sdk._resource import ResourceId
from ncae_sdk._util import map_async
from ncae_sdk.resources._schema import (
    BasePhase,
    BasePhaseFilter,
    BasePhaseFilterModel,
)


class AsyncBasePhaseEndpoint(AsyncEndpoint):
    BASE_PATH: Final[str] = "automation/v1/base-phase"

    async def list(self, **filters: Unpack[BasePhaseFilter]) -> AsyncIterator[BasePhase]:
        results = self._list(self.BASE_PATH, BasePhaseFilterModel, filters)
        return map_async(BasePhase.parse_api, results)

    async def find(self, **filters: Unpack[BasePhaseFilter]) -> Optional[BasePhase]:
        result = await self._get_by_filters(self.BASE_PATH, BasePhaseFilterModel, filters)
        return BasePhase.parse_api(result) if result else None

    async def get(self, rid: ResourceId) -> Optional[BasePhase]:
        result = await self._get_by_id(self.BASE_PATH, rid)
        return BasePhase.parse_api(result) if result else None



import logging
import logging.config
from typing import TYPE_CHECKING, Any, Callable, MutableMapping, Union

from typing_extensions import TypeAlias


def stdout_filter() -> Callable[[logging.LogRecord], bool]:
    return lambda record: record.levelno <= logging.INFO


def stderr_filter() -> Callable[[logging.LogRecord], bool]:
    return lambda record: record.levelno > logging.INFO


NCAE_LOGGING_CONFIG: dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": "%(asctime)s %(name)s: [%(levelname)-8s] %(message)s"},
    },
    "filters": {
        "stdout_filter": {"()": stdout_filter},
        "stderr_filter": {"()": stderr_filter},
    },
    "handlers": {
        "stdout": {
            "level": "DEBUG",
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "filters": ["stdout_filter"],
        },
        "stderr": {
            "level": "DEBUG",
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
            "filters": ["stderr_filter"],
        },
    },
    "root": {
        "handlers": ["stdout", "stderr"],
        "level": "INFO",
    },
    "loggers": {
        "asyncio": {"level": "WARNING"},
        "httpx": {"level": "WARNING"},
        "httpcore": {"level": "WARNING"},
        "uvicorn": {"level": "INFO"},
    },
}


def setup_default_logging(debug_mode: bool) -> None:
    config = NCAE_LOGGING_CONFIG.copy()
    if debug_mode:
        config["root"]["level"] = "DEBUG"

    logging.config.dictConfig(config)


# Workaround for proper type hinting of LoggerAdapter, as it varies between Python versions
# See https://github.com/python/typeshed/issues/7855 for more details
if TYPE_CHECKING:
    LoggerAdapter = logging.LoggerAdapter[logging.Logger]
else:
    LoggerAdapter = logging.LoggerAdapter

ContextLogger: TypeAlias = Union[logging.Logger, LoggerAdapter]


class ContextLogAdapter(LoggerAdapter):
    def process(self, msg: Any, kwargs: MutableMapping[str, Any]) -> tuple[str, MutableMapping[str, Any]]:
        assert isinstance(self.extra, dict)
        return "[%s] %s" % (self.extra["identifier"], msg), kwargs
